/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program and library             */
/*         SCIP --- Solving Constraint Integer Programs                      */
/*                                                                           */
/*    Copyright (C) 2002-2015 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SCIP is distributed under the terms of the ZIB Academic License.         */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with SCIP; see the file COPYING. If not email to scip@zib.de.      */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   cons_nosubtour.c
 * @brief  constraint handler for subtour elimination constraints
 * @author Felipe Serrano
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include <assert.h>
#include <string.h>

#include "cons_nosubtour.h"
#include "probdata_tsp.h"
#include "GomoryHuTree.h"


/**@name Constraint handler properties
 *
 * @{
 */

/* fundamental constraint handler properties */
#define CONSHDLR_NAME          "nosubtour"
#define CONSHDLR_DESC          "TSP subtour elimination constraints"
#define CONSHDLR_ENFOPRIORITY  -2000000 /**< priority of the constraint handler for constraint enforcing */
#define CONSHDLR_CHECKPRIORITY -2000000 /**< priority of the constraint handler for checking feasibility */
#define CONSHDLR_EAGERFREQ            1 /**< frequency for using all instead of only the useful constraints in separation,
                                              *   propagation and enforcement, -1 for no eager evaluations, 0 for first only */
#define CONSHDLR_NEEDSCONS         TRUE /**< should the constraint handler be skipped, if no constraints are available? */

/* optional constraint handler properties */
#define CONSHDLR_SEPAPRIORITY   1000000 /**< priority of the constraint handler for separation */
#define CONSHDLR_SEPAFREQ             1 /**< frequency for separating cuts; zero means to separate only in the root node */
#define CONSHDLR_DELAYSEPA        FALSE /**< should separation method be delayed, if other separators found cuts? */

#define CONSHDLR_PROPFREQ            -1 /**< frequency for propagating domains; zero means only preprocessing propagation */
#define CONSHDLR_DELAYPROP        FALSE /**< should propagation method be delayed, if other propagators found reductions? */
#define CONSHDLR_PROP_TIMING      SCIP_PROPTIMING_BEFORELP/**< propagation timing mask of the constraint handler*/

#define CONSHDLR_PRESOLTIMING     SCIP_PRESOLTIMING_FAST /**< presolving timing of the constraint handler (fast, medium, or exhaustive) */
#define CONSHDLR_MAXPREROUNDS        0 /**< maximal number of presolving rounds the constraint handler participates in (-1: no limit) */
/**@} */


/*
 * Data structures
 */

/** constraint data for nosubtour constraints */
struct SCIP_ConsData
{
   GRAPH* graph;
};

/**@name Local methods
 *
 * @{
 */


/* checks whether proposed solution contains a subtour */
static
SCIP_Bool findSubtour(
   SCIP*              scip,               /**< SCIP data structure */
   GRAPH*             graph,              /**< underlying graph */
   SCIP_SOL*          sol                 /**< proposed solution */
   )
{
   GRAPHNODE* node;
   GRAPHNODE* startnode;
   GRAPHEDGE* lastedge;
   GRAPHEDGE* edge;
   int tourlength;

   if(graph->nnodes <= 1)
      return FALSE;

   startnode = &graph->nodes[0];

   tourlength = 0;
   lastedge = NULL;
   node = startnode;

   /* follow the (sub?)tour until you come back to the startnode */
   do
   {
      edge = node->first_edge;

      /* look for an outgoing edge to proceed */
      while( edge->back == lastedge || SCIPgetSolVal(scip, sol, edge->var) < 0.5 )
      {
         edge = edge->next;
         assert(edge != NULL);
      }

      tourlength++;
      node = edge->adjac;
      lastedge = edge;
   }
   while( node != startnode );

   assert(tourlength <= graph->nnodes);

   return ( graph->nnodes != tourlength );
}

/* separates subtour elemination cuts */
static
SCIP_RETCODE sepaSubtour(
   SCIP*              scip,               /**< SCIP data structure */
   SCIP_CONSHDLR*     conshdlr,           /**< the constraint handler itself */
   SCIP_CONS**        conss,              /**< array of constraints to process */
   int                nconss,             /**< number of constraints to process */
   int                nusefulconss,       /**< number of useful (non-obsolete) constraints to process */
   SCIP_SOL*          sol,                /**< primal solution that should be separated */
   SCIP_RESULT*       result              /**< pointer to store the result of the separation call */
   )
{
   int c;
   int i;
   int j;

   assert(result != NULL);

   *result = SCIP_DIDNOTFIND;

   for( c = 0; c < nusefulconss && *result != SCIP_CUTOFF; ++c )
   {
      /* get all required structures */
      SCIP_CONSDATA* consdata;
      GRAPH* graph;
      SCIP_Bool** cuts;
      double cap;
      int ncuts;

      consdata = SCIPconsGetData(conss[c]);
      assert(consdata != NULL);

      graph = consdata->graph;
      assert(graph != NULL);

      /* store the suggested, but infeasible solution into the capacity of the edges */
      for( i = 0; i < graph->nedges; ++i )
      {
         cap = SCIPgetSolVal(scip, sol, graph->edges[i].var);
         graph->edges[i].rcap = cap;
         graph->edges[i].cap = cap;
         graph->edges[i].back->rcap = cap;
         graph->edges[i].back->cap = cap;
      }

      /* allocate two-dimensional array cuts of size nnodes*nnodes */
      SCIP_CALL( SCIPallocBufferArray(scip, &cuts, graph->nnodes) );
      for( i = 0; i < graph->nnodes; ++i )
      {
         SCIP_CALL( SCIPallocClearBufferArray(scip, &cuts[i], graph->nnodes) );
      }

      /* try to find cuts */
      if( ghc_tree(graph, cuts, &ncuts, SCIPfeastol(scip)) )
      {
         /* create a new cutting plane for every suitable arc (representing a cut with value < 2) of the Gomory Hu Tree */
         for( i = 0; i < ncuts && *result != SCIP_CUTOFF; ++i )
         {
            SCIP_ROW* row;
            SCIP_Bool infeasible;

            SCIP_CALL( SCIPcreateEmptyRowCons(scip, &row, conshdlr, "sepa_cons", 2.0, SCIPinfinity(scip), FALSE, FALSE, TRUE) );

            /* since we are adding variables with SCIPaddVarToRow(), we cache row extensions for performance reasons */
            SCIP_CALL( SCIPcacheRowExtensions(scip, row) );

            for( j = 0; j < graph->nnodes; ++j )
            {
               /* in gmincut the graph has been partitioned into two parts, represented by bools */
               if( cuts[i][j] )
               {
                  GRAPHEDGE* edge = graph->nodes[j].first_edge;

                  /* take every edge with nodes in different parts into account */
                  while( edge != NULL )
                  {
                     if( !cuts[i][edge->adjac->id] )
                     {
                        SCIP_CALL( SCIPaddVarToRow(scip, row, edge->var, 1.0) );
                     }
                     edge = edge->next;
                  }
               }
            }

            /* finalize the row by flushing all added coefficients; mandatory when using SCIPcacheRowExtensions() */
            SCIP_CALL( SCIPflushRowExtensions(scip, row) );

            /* add cut */
            SCIP_CALL( SCIPaddCut(scip, sol, row, FALSE, &infeasible) );
            if ( infeasible )
               *result = SCIP_CUTOFF;
            else
               *result = SCIP_SEPARATED;

            /* tells SCIP that the row pointer is no longer used by this constraint handler */
            SCIP_CALL( SCIPreleaseRow(scip, &row) );
         }
      }

      /* free memory in reverse order of allocation for performance reasons */
      for( i = graph->nnodes - 1; i >= 0; i-- )
         SCIPfreeBufferArray(scip, &cuts[i]);

      SCIPfreeBufferArray(scip, &cuts);
   }

   return SCIP_OKAY;
}

/**@} */


/**@name Callback methods
 *
 * @{
 */


/** copy method for constraint handler plugins (called when SCIP copies plugins) */
   static
SCIP_DECL_CONSHDLRCOPY(conshdlrCopyNosubtour)
{  /*lint --e{715}*/
   assert(scip != NULL);
   assert(conshdlr != NULL);
   assert(strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0);

   /* call inclusion method of constraint handler */
   SCIP_CALL( SCIPincludeConshdlrNosubtour(scip) );

   *valid = TRUE;

   return SCIP_OKAY;
}


/** frees specific constraint data */
static
SCIP_DECL_CONSDELETE(consDeleteNosubtour)
{  /*lint --e{715}*/
   assert(consdata != NULL);

   release_graph(&(*consdata)->graph);
   SCIPfreeMemory(scip, consdata);

   return SCIP_OKAY;
}


/** transforms constraint data into data belonging to the transformed problem */
static
SCIP_DECL_CONSTRANS(consTransNosubtour)
{  /*lint --e{715}*/
   SCIP_CONSDATA* sourcedata;
   SCIP_CONSDATA* targetdata;

   sourcedata = SCIPconsGetData(sourcecons);
   assert( sourcedata != NULL );
   targetdata = NULL;

   SCIP_CALL( SCIPallocMemory(scip, &targetdata) );
   targetdata->graph = sourcedata->graph;
   capture_graph(targetdata->graph);

   /* create target constraint */
   SCIP_CALL( SCIPcreateCons(scip, targetcons, SCIPconsGetName(sourcecons), conshdlr, targetdata,
         SCIPconsIsInitial(sourcecons), SCIPconsIsSeparated(sourcecons), SCIPconsIsEnforced(sourcecons),
         SCIPconsIsChecked(sourcecons), SCIPconsIsPropagated(sourcecons),  SCIPconsIsLocal(sourcecons),
         SCIPconsIsModifiable(sourcecons), SCIPconsIsDynamic(sourcecons), SCIPconsIsRemovable(sourcecons),
         SCIPconsIsStickingAtNode(sourcecons)) );

   return SCIP_OKAY;
}


/** separation method of constraint handler for LP solutions */
static
SCIP_DECL_CONSSEPALP(consSepalpNosubtour)
{  /*lint --e{715}*/
   SCIP_CALL( sepaSubtour(scip, conshdlr, conss, nconss, nusefulconss, NULL, result) );

   return SCIP_OKAY;
}


/** separation method of constraint handler for arbitrary primal solutions */
static
SCIP_DECL_CONSSEPASOL(consSepasolNosubtour)
{  /*lint --e{715}*/
   SCIP_CALL( sepaSubtour(scip, conshdlr, conss, nconss, nusefulconss, sol, result) );

   return SCIP_OKAY;
}


/** constraint enforcing method of constraint handler for LP solutions */
static
SCIP_DECL_CONSENFOLP(consEnfolpNosubtour)
{  /*lint --e{715}*/
   int i;

   /* the subtour elimination constraints are enforced after the node degree and integrality constraints.
    * we will only check feasibility, when the others are satisfied.
    */
   if( solinfeasible )
   {
      *result = SCIP_INFEASIBLE;
      return SCIP_OKAY;
   }

   *result = SCIP_FEASIBLE;
   for( i = 0; i < nconss; ++i )
   {
      SCIP_CONSDATA* consdata;
      GRAPH* graph;
      SCIP_Bool found;

      consdata = SCIPconsGetData(conss[i]);
      assert(consdata != NULL);

      graph = consdata->graph;
      assert(graph != NULL);

      found = findSubtour(scip, graph, NULL);

      /* if a subtour is found, the solution must be infeasible */
      if( found )
         *result = SCIP_INFEASIBLE;
   }

   return SCIP_OKAY;
}


/** constraint enforcing method of constraint handler for pseudo solutions */
static
SCIP_DECL_CONSENFOPS(consEnfopsNosubtour)
{  /*lint --e{715}*/
   int i;

   /* the subtour elimination constraints are enforced after the node degree and integrality constraints.
    * we will only check feasibility, when the others are satisfied.
    */
   if( solinfeasible )
   {
      *result = SCIP_INFEASIBLE;
      return SCIP_OKAY;
   }

   *result = SCIP_FEASIBLE;
   for( i = 0; i < nconss; ++i )
   {
      SCIP_CONSDATA* consdata;
      GRAPH* graph;
      SCIP_Bool found;

      consdata = SCIPconsGetData(conss[i]);
      assert(consdata != NULL);
      graph = consdata->graph;
      assert(graph != NULL);

      /* if a subtour is found, the solution must be infeasible */
      found = findSubtour(scip, graph, NULL);
      if( found )
         *result = SCIP_INFEASIBLE;
   }

   return SCIP_OKAY;
}


/** feasibility check method of constraint handler for integral solutions */
static
SCIP_DECL_CONSCHECK(consCheckNosubtour)
{  /*lint --e{715}*/
   int i;

   *result = SCIP_FEASIBLE;

   for( i = 0; i < nconss; ++i )
   {
      SCIP_CONSDATA* consdata;
      GRAPH* graph;
      SCIP_Bool found;

      consdata = SCIPconsGetData(conss[i]);
      assert(consdata != NULL);
      graph = consdata->graph;
      assert(graph != NULL);

      /* the subtour elimination constraints are checked after the node degree and integrality constraints.
       * if one of them was infeasible, we would not have been called
       */
      found = findSubtour(scip, graph, sol);

      /* if a subtour is found, the solution must be infeasible */
      if( found )
      {
         *result = SCIP_INFEASIBLE;
         if( printreason )
         {
            SCIP_CALL( SCIPprintCons(scip, conss[i], NULL) );
            SCIPinfoMessage(scip, NULL, "violation: graph has a subtour\n");
         }
      }
   }

   return SCIP_OKAY;
}


/** domain propagation method of constraint handler */
static
SCIP_DECL_CONSPROP(consPropNosubtour)
{  /*lint --e{715}*/
   assert(result != NULL);
   *result = SCIP_DIDNOTRUN;
   return SCIP_OKAY;
}


/** variable rounding lock method of constraint handler */
static
SCIP_DECL_CONSLOCK(consLockNosubtour)
{  /*lint --e{715}*/
   SCIP_CONSDATA* consdata;
   GRAPH* g;
   int i;

   consdata = SCIPconsGetData(cons);
   assert(consdata != NULL);

   g = consdata->graph;
   assert(g != NULL);

   for( i = 0; i < g->nedges; ++i )
   {
      SCIP_CALL( SCIPaddVarLocks(scip, g->edges[i].var, nlocksneg, nlockspos) );
   }

   return SCIP_OKAY;
}


/** variable deletion of constraint handler */
static
SCIP_DECL_CONSDELVARS(consDelvarsNosubtour)
{  /*lint --e{715}*/
   return SCIP_OKAY;
}


/** constraint display method of constraint handler */
static
SCIP_DECL_CONSPRINT(consPrintNosubtour)
{  /*lint --e{715}*/
   SCIP_CONSDATA* consdata;
   GRAPH* g;

   consdata = SCIPconsGetData(cons);
   assert(consdata != NULL);

   g = consdata->graph;
   assert(g != NULL);

   SCIPinfoMessage(scip, file, "subtour of Graph G with %d nodes and %d edges\n", g->nnodes, g->nedges);

   return SCIP_OKAY;
}


/** constraint copying method of constraint handler */
static
SCIP_DECL_CONSCOPY(consCopyNosubtour)
{  /*lint --e{715}*/
   SCIP_PROBDATA* probdata;
   GRAPH* graph;

   probdata = SCIPgetProbData(scip);
   assert(probdata != NULL);

   graph = SCIPprobdataGetGraph(probdata);

   /* create constraint */
   SCIP_CALL( SCIPcreateConsBasicNosubtour(scip, cons, (name == NULL) ? SCIPconsGetName(sourcecons) : name,  graph) );

   *valid = TRUE;
   return SCIP_OKAY;
}

/**@} */


/**@name Interface methods
 *
 * @{
 */


/** creates the handler for nosubtour constraints and includes it in SCIP */
SCIP_RETCODE SCIPincludeConshdlrNosubtour(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_CONSHDLRDATA* conshdlrdata;
   SCIP_CONSHDLR* conshdlr;

   /* create nosubtour constraint handler data */
   conshdlrdata = NULL;
   /* TODO: (optional) create constraint handler specific data here */

   conshdlr = NULL;

   /* include constraint handler */
   SCIP_CALL( SCIPincludeConshdlrBasic(scip, &conshdlr, CONSHDLR_NAME, CONSHDLR_DESC,
         CONSHDLR_ENFOPRIORITY, CONSHDLR_CHECKPRIORITY, CONSHDLR_EAGERFREQ, CONSHDLR_NEEDSCONS,
         consEnfolpNosubtour, consEnfopsNosubtour, consCheckNosubtour, consLockNosubtour,
         conshdlrdata) );
   assert(conshdlr != NULL);

   /* set non-fundamental callbacks via specific setter functions */
   SCIP_CALL( SCIPsetConshdlrCopy(scip, conshdlr, conshdlrCopyNosubtour, consCopyNosubtour) );
   SCIP_CALL( SCIPsetConshdlrDelete(scip, conshdlr, consDeleteNosubtour) );
   SCIP_CALL( SCIPsetConshdlrDelvars(scip, conshdlr, consDelvarsNosubtour) );
   SCIP_CALL( SCIPsetConshdlrPrint(scip, conshdlr, consPrintNosubtour) );
   SCIP_CALL( SCIPsetConshdlrProp(scip, conshdlr, consPropNosubtour, CONSHDLR_PROPFREQ, CONSHDLR_DELAYPROP,
         CONSHDLR_PROP_TIMING) );
   SCIP_CALL( SCIPsetConshdlrSepa(scip, conshdlr, consSepalpNosubtour, consSepasolNosubtour, CONSHDLR_SEPAFREQ, CONSHDLR_SEPAPRIORITY, CONSHDLR_DELAYSEPA) );
   SCIP_CALL( SCIPsetConshdlrTrans(scip, conshdlr, consTransNosubtour) );

   return SCIP_OKAY;
}

/** creates and captures a nosubtour constraint
 *
 *  @note the constraint gets captured, hence at one point you have to release it using the method SCIPreleaseCons()
 */
SCIP_RETCODE SCIPcreateConsNosubtour(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_CONS**           cons,               /**< pointer to hold the created constraint */
   const char*           name,               /**< name of constraint */
   GRAPH*                graph,              /**< the underlying graph */
   SCIP_Bool             initial,            /**< should the LP relaxation of constraint be in the initial LP?
                                              *   Usually set to TRUE. Set to FALSE for 'lazy constraints'. */
   SCIP_Bool             separate,           /**< should the constraint be separated during LP processing?
                                              *   Usually set to TRUE. */
   SCIP_Bool             enforce,            /**< should the constraint be enforced during node processing?
                                              *   TRUE for model constraints, FALSE for additional, redundant constraints. */
   SCIP_Bool             check,              /**< should the constraint be checked for feasibility?
                                              *   TRUE for model constraints, FALSE for additional, redundant constraints. */
   SCIP_Bool             propagate,          /**< should the constraint be propagated during node processing?
                                              *   Usually set to TRUE. */
   SCIP_Bool             local,              /**< is constraint only valid locally?
                                              *   Usually set to FALSE. Has to be set to TRUE, e.g., for branching constraints. */
   SCIP_Bool             modifiable,         /**< is constraint modifiable (subject to column generation)?
                                              *   Usually set to FALSE. In column generation applications, set to TRUE if pricing
                                              *   adds coefficients to this constraint. */
   SCIP_Bool             dynamic,            /**< is constraint subject to aging?
                                              *   Usually set to FALSE. Set to TRUE for own cuts which
                                              *   are separated as constraints. */
   SCIP_Bool             removable,          /**< should the relaxation be removed from the LP due to aging or cleanup?
                                              *   Usually set to FALSE. Set to TRUE for 'lazy constraints' and 'user cuts'. */
   SCIP_Bool             stickingatnode      /**< should the constraint always be kept at the node where it was added, even
                                              *   if it may be moved to a more global node?
                                              *   Usually set to FALSE. Set to TRUE to for constraints that represent node data. */
   )
{
   SCIP_CONSHDLR* conshdlr;
   SCIP_CONSDATA* consdata;

   /* find the nosubtour constraint handler */
   conshdlr = SCIPfindConshdlr(scip, CONSHDLR_NAME);
   if( conshdlr == NULL )
   {
      SCIPerrorMessage("nosubtour constraint handler not found\n");
      return SCIP_PLUGINNOTFOUND;
   }

   /* create constraint data */
   SCIP_CALL( SCIPallocMemory( scip, &consdata) );
   consdata->graph = graph;
   capture_graph(consdata->graph);

   /* create constraint */
   SCIP_CALL( SCIPcreateCons(scip, cons, name, conshdlr, consdata, initial, separate, enforce, check, propagate,
         local, modifiable, dynamic, removable, stickingatnode) );

   return SCIP_OKAY;
}

/** creates and captures a nosubtour constraint with all its constraint flags set to their
 *  default values
 *
 *  @note the constraint gets captured, hence at one point you have to release it using the method SCIPreleaseCons()
 */
SCIP_RETCODE SCIPcreateConsBasicNosubtour(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_CONS**           cons,               /**< pointer to hold the created constraint */
   const char*           name,               /**< name of constraint */
   GRAPH*                graph               /**< the underlying graph */
   )
{
   SCIP_CALL( SCIPcreateConsNosubtour(scip, cons, name, graph,
         TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE) );

   return SCIP_OKAY;
}
