/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program and library             */
/*         SCIP --- Solving Constraint Integer Programs                      */
/*                                                                           */
/*    Copyright (C) 2002-2015 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SCIP is distributed under the terms of the ZIB Academic License.         */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License.             */
/*  along with SCIP; see the file COPYING. If not email to scip@zib.de.      */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   event_newsol.c
 * @brief  event handler for new solutions in TSP
 * @author Timo Berthold
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include <assert.h>
#ifndef _MSC_VER
#include <unistd.h>
#else
#include <windows.h>
#define sleep Sleep
#endif

#include "scip/scip.h"
#include "event_newsol.h"
#include "probdata_tsp.h"
#include "GomoryHuTree.h"

#define EVENTHDLR_NAME         "newsol"
#define EVENTHDLR_DESC         "event handler for new solution in TSP"


/** initialization method of event handler (called after problem was transformed) */
static
SCIP_DECL_EVENTINIT(eventInitNewsol)
{
   FILE* lockfile;
   FILE* filedata;
   int lockwaits;

   lockwaits = 0;
   while( SCIPfileExists("temp.tour.lock") && lockwaits < 10 )
   {
      /* wait one second and try again */
      (void) sleep(1);
      lockwaits++;
   }

   if( SCIPfileExists("temp.tour.lock") )
   {
      SCIPwarningMessage(scip, "cannot reset, because lockfile <temp.tour.lock> is still existing\n");
      return SCIP_OKAY;
   }

   /* create lock file */
   lockfile = fopen("temp.tour.lock", "w");
   if( lockfile == NULL )
   {
      SCIPerrorMessage("cannot create file temp.tour.lock for writing\n");
      return SCIP_FILECREATEERROR;
   }
   
   fprintf(lockfile, "lock\n");
   fclose(lockfile);

   /* create output file which can be read by TSPViewer */
   filedata = fopen("temp.tour", "w");
   if( filedata == NULL )
   {
      SCIPerrorMessage("cannot create file temp.tour for writing\n");
      return SCIP_FILECREATEERROR;
   }

   fprintf(filedata, "RESET\n");
   fclose(filedata);

   /* delete lock file */
   unlink("temp.tour.lock");
   (void) sleep(1); /* wait for the Java TSPViewer */

   return SCIP_OKAY;
} /*lint !e715*/


/** solving process initialization method of event handler (called when branch and bound process is about to begin)
 *
 *  This method is called when the presolving was finished and the branch and bound process is about to begin.
 *  The event handler may use this call to initialize its branch and bound specific data.
 *
 */
static
SCIP_DECL_EVENTINITSOL(eventInitsolNewsol)
{
   SCIP_CALL( SCIPcatchEvent(scip, SCIP_EVENTTYPE_BESTSOLFOUND, eventhdlr, NULL, NULL) );
   return SCIP_OKAY;
}


/** solving process deinitialization method of event handler (called before branch and bound process data is freed)
 *
 *  This method is called before the branch and bound process is freed.
 *  The event handler should use this call to clean up its branch and bound data.
 */
static
SCIP_DECL_EVENTEXITSOL(eventExitsolNewsol)
{
   SCIP_CALL( SCIPdropEvent(scip, SCIP_EVENTTYPE_BESTSOLFOUND, eventhdlr, NULL, -1) );
   return SCIP_OKAY;
}


/** execution method of event handler
 *
 *  Processes the event. The method is called every time an event occurs, for which the event handler
 *  is responsible. Event handlers may declare themselves resposible for events by calling the
 *  corresponding SCIPcatch...() method. This method creates an event filter object to point to the
 *  given event handler and event data.
 */
static
SCIP_DECL_EVENTEXEC(eventExecNewsol)
{
   SCIP_SOL* sol;
   SCIP_HEUR* heur;
   SCIP_PROBDATA* probdata;
   GRAPH* graph;
   GRAPHNODE* node;
   GRAPHEDGE* lastedge;
   FILE* lockfile;
   FILE* filedata;
   int lockwaits;

   probdata = SCIPgetProbData(scip);
   assert(probdata != NULL);

   graph = SCIPprobdataGetGraph(probdata);
   node = &graph->nodes[0];

   lastedge = NULL;
   lockwaits = 0;
   sol = SCIPgetBestSol(scip);

   /* wait for lock file to disappear */
   while( SCIPfileExists("temp.tour.lock") && lockwaits < 10 )
   {
      /* wait one second and try again */
      (void) sleep(1);
      lockwaits++;
   }

   if( SCIPfileExists("temp.tour.lock") )
   {
      SCIPwarningMessage(scip, "cannot output tour in file, because lockfile <temp.tour.lock> is still existing\n");
      return SCIP_OKAY;
   }

   /* create lock file */
   lockfile = fopen("temp.tour.lock", "w");
   if( lockfile == NULL )
   {
      SCIPerrorMessage("cannot create file temp.tour.lock for writing\n");
      return SCIP_FILECREATEERROR;
   }
   fprintf(lockfile, "lock\n");
   fclose(lockfile);

   /* create output file which can be read by TSPViewer */
   filedata = fopen("temp.tour", "w");
   if( filedata == NULL )
   {
      SCIPerrorMessage("cannot create file temp.tour for writing\n");
      return SCIP_FILECREATEERROR;
   }
   fprintf(filedata, "%d\n", graph->nnodes);

   heur = SCIPgetSolHeur(scip, sol);
   if ( heur == NULL)
      fprintf(filedata, "relaxation\n");
   else
      fprintf(filedata, "%s\n", SCIPheurGetName(heur));

   fprintf(filedata, "%g\n", SCIPgetSolOrigObj(scip, sol));
   do
   {
      GRAPHEDGE* edge;

      edge= node->first_edge;

      /* output the number of nodes */
      fprintf(filedata, "%d %g %g\n", node->id, node->x, node->y);

      /* output the id and the coordinates of the nodes in the order they appear in the tour */
      while( edge != NULL)
      {
         if( edge->back != lastedge && SCIPgetSolVal(scip, sol, edge->var) > 0.5 )
         {
            node = edge->adjac;
            lastedge = edge;
            break;
         }
         edge = edge->next;
      }
   }
   while ( node != &graph->nodes[0] );

   fclose(filedata);

   /* delete lock file */
   unlink("temp.tour.lock");
   (void) sleep(1); /* wait for the Java TSPViewer */

   return SCIP_OKAY;
} /*lint !e715*/


/*
 * Interface methods
 */

/** includes event handler for new solution */
SCIP_RETCODE SCIPincludeEventHdlrNewsol(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_EVENTHDLR* eventhdlr;

   eventhdlr = NULL;

   /* include event handler into SCIP */
   SCIP_CALL( SCIPincludeEventhdlrBasic(scip, &eventhdlr, EVENTHDLR_NAME, EVENTHDLR_DESC,
            eventExecNewsol, NULL) );
   assert(eventhdlr != NULL);

   /* set non fundamental callbacks via setter functions */
   SCIP_CALL( SCIPsetEventhdlrInit(scip, eventhdlr, eventInitNewsol) );
   SCIP_CALL( SCIPsetEventhdlrInitsol(scip, eventhdlr, eventInitsolNewsol) );
   SCIP_CALL( SCIPsetEventhdlrExitsol(scip, eventhdlr, eventExitsolNewsol) );

   return SCIP_OKAY;
}
