/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program and library             */
/*         SCIP --- Solving Constraint Integer Programs                      */
/*                                                                           */
/*    Copyright (C) 2002-2015 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SCIP is distributed under the terms of the ZIB Academic License.         */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License.             */
/*  along with SCIP; see the file COPYING. If not email to scip@zib.de.      */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   heur_2opt.c
 * @brief  2-Optimum - combinatorial improvement heuristic for TSP
 * @author Timo Berthold
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include <assert.h>
#include <string.h>

#include "scip/scip.h"
#include "GomoryHuTree.h"
#include "heur_2opt.h"
#include "probdata_tsp.h"

#define HEUR_NAME             "2 opt"
#define HEUR_DESC             "2-Opt heuristic for TSPs"
#define HEUR_DISPCHAR         'K'
#define HEUR_PRIORITY         -1000000
#define HEUR_FREQ             1
#define HEUR_FREQOFS          0
#define HEUR_MAXDEPTH         -1
#define HEUR_TIMING           SCIP_HEURTIMING_AFTERNODE
#define HEUR_USESSUBSCIP      FALSE  /**< does the heuristic use a secondary SCIP instance? */

/** primal heuristic data */
struct SCIP_HeurData
{
   GRAPH*                graph;              /**< the underlying graph of the TSP */
   int                   ncalls;             /**< number of calls of the heuristic since the last solution was found */
   SCIP_SOL*             sol;                /**< current solution */
   GRAPHEDGE**           tour;               /**< tour induced by sol */
};

/*
 * Local methods
 */

/** method finding the edge going from the node with id index1 to the node with id index2 */
static
GRAPHEDGE* findEdge(
   GRAPHNODE*         nodes,              /**< all nodes of the graph */
   GRAPHNODE*         node1,              /**< id of the node where the searched edge starts */
   GRAPHNODE*         node2               /**< id of the node where the searched edge ends */
   )
{
   GRAPHEDGE* edge =  node1->first_edge;

   /* regard every outgoing edge of node index1 and stop if adjacent to node index2 */
   while( edge != NULL )
   {
      if( edge->adjac == node2 )
         break;

      edge = edge->next;
   }

   return edge;   
} /*lint !e715*/


/*
 * Callback methods of primal heuristic
 */

/** copy method for primal heuristic plugins (called when SCIP copies plugins) */
static
SCIP_DECL_HEURCOPY(heurCopy2opt)
{  /*lint --e{715}*/
   assert(scip != NULL);
   assert(heur != NULL);
   assert(strcmp(SCIPheurGetName(heur), HEUR_NAME) == 0);

   /* call inclusion method of primal heuristic */
   SCIP_CALL( SCIPincludeHeur2opt(scip) );

   return SCIP_OKAY;
}

/** solving process initialization method of primal heuristic (called when branch and bound process is about to begin)
 *
 *  This method is called when the presolving was finished and the branch and bound process is about to begin.
 *  The primal heuristic may use this call to initialize its branch and bound specific data.
 *
 */
static
SCIP_DECL_HEURINITSOL(heurInitsol2opt)
{
   SCIP_HEURDATA* heurdata;
   SCIP_PROBDATA* probdata;

   assert(strcmp(SCIPheurGetName(heur), HEUR_NAME) == 0);
   assert(scip != NULL);
   assert(heur != NULL);

   heurdata = SCIPheurGetData(heur);
   assert(heurdata != NULL);

   probdata = SCIPgetProbData(scip);
   assert(probdata != NULL);

   heurdata->graph = SCIPprobdataGetGraph(probdata);
   capture_graph(heurdata->graph);

   heurdata->ncalls = 0;
   heurdata->sol = NULL;
   SCIP_CALL( SCIPallocMemoryArray(scip, &heurdata->tour, heurdata->graph->nnodes) );
   
   return SCIP_OKAY;
} /*lint !e715*/

   
/** solving process deinitialization method of primal heuristic (called before branch and bound process data is freed)
 *
 *  This method is called before the branch and bound process is freed.
 *  The primal heuristic should use this call to clean up its branch and bound data.
 */
static
SCIP_DECL_HEUREXITSOL(heurExitsol2opt)
{
   SCIP_HEURDATA* heurdata;

   assert(strcmp(SCIPheurGetName(heur), HEUR_NAME) == 0);
   assert(scip != NULL);
   assert(heur != NULL);

   heurdata = SCIPheurGetData(heur);
   assert(heurdata != NULL);

   release_graph(&heurdata->graph);
   SCIPfreeMemoryArray(scip, &heurdata->tour);

   return SCIP_OKAY;
} /*lint !e715*/

/** destructor of primal heuristic to free user data (called when SCIP is exiting) */
static
SCIP_DECL_HEURFREE(heurFree2opt)
{
   SCIP_HEURDATA* heurdata;

   assert(strcmp(SCIPheurGetName(heur), HEUR_NAME) == 0);
   assert(scip != NULL);
   assert(heur != NULL);

   heurdata = SCIPheurGetData(heur);
   assert(heurdata != NULL);

   SCIPfreeMemory(scip, &heurdata);

   return SCIP_OKAY;
}


/** execution method of primal heuristic 2-Opt */
static
SCIP_DECL_HEUREXEC(heurExec2opt)
{
   SCIP_HEURDATA* heurdata;
   SCIP_SOL* sol;
   SCIP_Bool newsol;
   GRAPHNODE* nodes;
   GRAPHEDGE** edges2test;
   int nnodes;
   int i;
   int j;

   assert(strcmp(SCIPheurGetName(heur), HEUR_NAME) == 0);
   assert(scip != NULL);
   assert(heur != NULL);

   heurdata = SCIPheurGetData(heur);
   assert(heurdata != NULL);

   sol = SCIPgetBestSol( scip );

   /* check whether a new solution was found meanwhile */
   if( sol != heurdata->sol )
   {
      heurdata->sol = sol;
      heurdata->ncalls = 0;
      newsol = TRUE;
   }
   else
      newsol = FALSE;

   heurdata->ncalls++;

   nnodes = heurdata->graph->nnodes;

   /* some cases need not to be handled */
   if( nnodes < 4 || sol == NULL || heurdata->ncalls >= nnodes )
   {
      *result = SCIP_DIDNOTRUN;
      return SCIP_OKAY;
   }

   *result= SCIP_DIDNOTFIND;

   nodes = heurdata->graph->nodes;

   /* get tour from sol and sort edges by length, if new solution was found */
   if( newsol )
   {
      GRAPHEDGE* edge;
      GRAPHEDGE* lastedge = NULL;
      GRAPHNODE* node = &nodes[0];

      i = 0; 
      do
      {
         edge = node->first_edge;      
         while( edge != NULL )
         {
            /* find the next edge of the tour */
            if( edge->back != lastedge && SCIPgetSolVal(scip, sol, edge->var) > 0.5 )
            {
               node = edge->adjac;
               lastedge = edge;

               /* shift edge through the (sorted) array */
               for( j = i; j > 0 && heurdata->tour[j-1]->length < edge->length; j-- ) /*lint !e613*/
                  heurdata->tour[j] = heurdata->tour[j-1];
                               
               /* and insert the edge at the right position */
               heurdata->tour[j] = edge;

               i++;
               break;
            }
            edge = edge->next;

         }
      }while ( node != &nodes[0] );
      assert( i == nnodes );

   }

   SCIP_CALL( SCIPallocBufferArray(scip, &edges2test, 4) ); 

   /* test current edge with all 'longer' edges for improvement 
    * if swapping with crossing edges (though do 2Opt for one edge)
    */
   for( i = 0; i < heurdata->ncalls && *result != SCIP_FOUNDSOL; i++ )
   {
      edges2test[0] = heurdata->tour[heurdata->ncalls]; /*lint !e613*/
      edges2test[1] = heurdata->tour[i]; /*lint !e613*/
      edges2test[2] = findEdge( nodes, edges2test[0]->back->adjac, edges2test[1]->back->adjac );  
      edges2test[3] = findEdge( nodes, edges2test[0]->adjac, edges2test[1]->adjac );

      assert( edges2test[2] != NULL );
      assert( edges2test[3] != NULL );
             
      /* if the new solution is better and variables are not fixed, update and end */
      if( edges2test[0]->length + edges2test[1]->length > edges2test[2]->length + edges2test[3]->length 
         &&  SCIPvarGetLbGlobal(edges2test[0]->var) == 0.0
         &&  SCIPvarGetLbGlobal(edges2test[1]->var) == 0.0
         &&  SCIPvarGetUbGlobal(edges2test[2]->var) == 1.0
         &&  SCIPvarGetUbGlobal(edges2test[3]->var) == 1.0 )
      {
         SCIP_Bool success;
         SCIP_SOL* swapsol; /* copy of sol with 4 edges swapped */

         SCIP_CALL( SCIPcreateSol(scip, &swapsol, heur) );

         /* copy the old solution */
         for( j = 0; j < nnodes; j++)
         {
            SCIP_CALL( SCIPsetSolVal(scip, swapsol, heurdata->tour[j]->var, 1.0) );
         }

         /* and replace two edges */
         SCIP_CALL( SCIPsetSolVal(scip, swapsol, edges2test[0]->var, 0.0) );
         SCIP_CALL( SCIPsetSolVal(scip, swapsol, edges2test[1]->var, 0.0) );
         SCIP_CALL( SCIPsetSolVal(scip, swapsol, edges2test[2]->var, 1.0) );
         SCIP_CALL( SCIPsetSolVal(scip, swapsol, edges2test[3]->var, 1.0) );
         SCIP_CALL( SCIPaddSolFree(scip, &swapsol, &success) );

         assert(success);
         *result = SCIP_FOUNDSOL;                   
         heurdata->ncalls = 0;

      }
   }
   SCIPfreeBufferArray(scip, &edges2test);

   return SCIP_OKAY;
} /*lint !e715*/

/*
 * Primal heuristic specific interface methods
 */

/** creates the 2opt primal heuristic and includes it in SCIP */
SCIP_RETCODE SCIPincludeHeur2opt(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_HEURDATA* heurdata;
   SCIP_HEUR* heur;

   /* create 2opt primal heuristic data */
   SCIP_CALL( SCIPallocMemory(scip, &heurdata) );
   heur = NULL;

   /* include primal heuristic */
   SCIP_CALL( SCIPincludeHeurBasic(scip, &heur,
            HEUR_NAME, HEUR_DESC, HEUR_DISPCHAR, HEUR_PRIORITY, HEUR_FREQ, HEUR_FREQOFS,
            HEUR_MAXDEPTH, HEUR_TIMING, HEUR_USESSUBSCIP, heurExec2opt, heurdata) );
   assert(heur != NULL);

   /* set non fundamental callbacks via setter functions */
   SCIP_CALL( SCIPsetHeurCopy(scip, heur, heurCopy2opt) );
   SCIP_CALL( SCIPsetHeurInitsol(scip, heur, heurInitsol2opt) );
   SCIP_CALL( SCIPsetHeurExitsol(scip, heur, heurExitsol2opt) );
   SCIP_CALL( SCIPsetHeurFree(scip, heur, heurFree2opt) );

   return SCIP_OKAY;
}
