/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program and library             */
/*         SCIP --- Solving Constraint Integer Programs                      */
/*                                                                           */
/*    Copyright (C) 2002-2015 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SCIP is distributed under the terms of the ZIB Academic License.         */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License.             */
/*  along with SCIP; see the file COPYING. If not email to scip@zib.de.      */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   probdata_tsp.c
 * @brief  Problem data for TSP
 * @author Felipe Serrano
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "scip/scip.h"
#include "probdata_tsp.h"
#include "GomoryHuTree.h"

/** @brief Problem data which is accessible in all places
 *
 * This problem data is used to store the TSP graph
 */
struct SCIP_ProbData
{
   GRAPH*                graph;              /**< the graph */
};

/**@name Local methods
 *
 * @{
 */

/** copies given graph */
static
SCIP_RETCODE copy_graph(
   GRAPH* sourcegraph,                   /**< graph to be copied */
   GRAPH** graph                         /**< pointer to store the copied graph */
   )
{
   int n;
   int m;
   int i;
   int j;
   int e;

   assert( graph != NULL );
   assert( sourcegraph != NULL );

   /* copy graph the way it is created in the file reader */
   n = sourcegraph->nnodes;
   m = sourcegraph->nedges;

   /* create_graphs allocates memory for 2 anti-parallel arcs for each edge */
   if(!create_graph(n, 2*m, graph))
      return SCIP_NOMEMORY;

   /* copy nodes */
   for( i = 0; i < n; ++i )
   {
      GRAPHNODE* node       = &((*graph)->nodes[i]);
      GRAPHNODE* sourcenode = &(sourcegraph->nodes[i]);

      assert(sourcenode->id == i);

      node->x          = sourcenode->x;
      node->y          = sourcenode->y;
      node->id         = sourcenode->id;
      node->first_edge = NULL;
   }

   /* copy edges */
   e = 0;			     
   for( i = 0; i < n - 1; ++i )
   {
      GRAPHNODE * nodestart = &((*graph)->nodes[i]);
      for( j = i + 1; j < n; ++j )
      {
         GRAPHNODE * nodeend = &((*graph)->nodes[j]);

         GRAPHEDGE * edgeforw  = &((*graph)->edges[e]);
         GRAPHEDGE * edgebackw = &((*graph)->edges[e + m]);

         /* construct two 'parallel' halfedges */
         edgeforw->adjac  = nodeend;
         edgebackw->adjac = nodestart;
         edgeforw->back   = edgebackw;
         edgebackw->back  = edgeforw;

         /* copy length */
         edgeforw->length  = sourcegraph->edges[e].length;
         edgebackw->length = edgeforw->length;

         /* insert one of the halfedges into the edge list of the node */
         if (nodestart->first_edge == NULL)
         {
            nodestart->first_edge = edgeforw;
            nodestart->first_edge->next = NULL;
         }
         else
         {
            edgeforw->next = nodestart->first_edge;
            nodestart->first_edge = edgeforw;
         }

         /* dito */
         if (nodeend->first_edge == NULL)
         {
            nodeend->first_edge = edgebackw;
            nodeend->first_edge->next = NULL;
         }
         else
         {
            edgebackw->next = nodeend->first_edge;
            nodeend->first_edge = edgebackw;
         }           

         ++e;
      } /* for j */
   } /* for i */

   return SCIP_OKAY;
}


/**@} */

/**@name Callback methods of problem data
 *
 * @{
 */

/** copies user data of source SCIP for the target SCIP */
static
SCIP_DECL_PROBCOPY(probcopyTsp)
{
   GRAPH* graph;
   GRAPH* sourcegraph;
   int e;
   int m;

   assert(scip != NULL);
   assert(sourcescip != NULL);
   assert(sourcedata != NULL);

   sourcegraph = sourcedata->graph;
   assert(sourcegraph != NULL);

   /* copy graph */
   SCIP_CALL( copy_graph(sourcegraph, &graph) );

   /* copy and link variables */
   m = graph->nedges;
   for( e = 0; e < m; ++e )
   {
      SCIP_Bool success;
      GRAPHEDGE* edgeforw  = &(graph->edges[e]);
      GRAPHEDGE* edgebackw = &(graph->edges[e + m]);

      assert( sourcegraph->edges[e].var != NULL );

      SCIP_CALL( SCIPgetVarCopy(sourcescip, scip, sourcegraph->edges[e].var, &(edgeforw->var), varmap, consmap, global, &success) );
      assert(success);
      assert(edgeforw->var != NULL);

      /* anti-parallel arcs share variable */
      edgebackw->var = edgeforw->var;
      SCIP_CALL( SCIPcaptureVar(scip, edgebackw->var) );
   }

   /* allocate memory for target prob data */
   SCIP_CALL( SCIPallocMemory(scip, targetdata) );

   /* set graph and capture it */
   (*targetdata)->graph = graph;
   capture_graph(graph);
   
   /* graph is captured by targetdata */
   release_graph(&graph);
   
   *result = SCIP_SUCCESS;
   return SCIP_OKAY;
}


/** frees user data of original problem (called when the original problem is freed) */
static
SCIP_DECL_PROBDELORIG(probdelorigTsp)
{
   GRAPH* graph;
   int i;

   assert(probdata != NULL);
   assert(*probdata != NULL);

   graph = (*probdata)->graph;
   assert(graph != NULL);

   for( i = 0; i < graph->nedges; i++ )
   {
      SCIP_CALL( SCIPreleaseVar(scip, &graph->edges[i].var) );
   }
   release_graph(&graph);

   /* free probdata */
   SCIPfreeMemory(scip, probdata);

   return SCIP_OKAY;
}

/** creates user data of transformed problem by transforming the original user problem data
 *  (called after problem was transformed) */
static
SCIP_DECL_PROBTRANS(probtransTsp)
{
   GRAPH* transgraph;
   GRAPH* graph;
   int e;
   int m;

   assert( scip != NULL );
   assert( sourcedata != NULL );
   assert( targetdata != NULL );

   graph = sourcedata->graph;
   assert( graph != NULL );

   /* copy graph */
   transgraph = NULL;
   SCIP_CALL( copy_graph(graph, &transgraph) );

   /* copy and link variables */
   m = transgraph->nedges;
   for( e = 0; e < m; ++e )
   {
      GRAPHEDGE * edgeforw  = &(transgraph->edges[e]);
      GRAPHEDGE * edgebackw = &(transgraph->edges[e + m]);

      assert( graph->edges[e].var != NULL );

      SCIP_CALL( SCIPgetTransformedVar(scip, graph->edges[e].var, &(edgeforw->var)) );
      edgebackw->var = edgeforw->var; /* anti-parallel arcs share variable */
      assert( edgebackw->var != NULL );
      SCIP_CALL( SCIPcaptureVar(scip, edgebackw->var) );
   }

   /* allocate memory for target prob data */
   SCIP_CALL( SCIPallocMemory(scip, targetdata) );

   /* set graph and capture it */
   (*targetdata)->graph = transgraph;
   capture_graph(transgraph);

   /* graph is captured by targetdata */
   release_graph(&transgraph);
   
   return SCIP_OKAY;
}

#if 0
static
SCIP_DECL_PROBEXITSOL(probexitsolTsp)
{  /*lint --e{715}*/
   if( graph_ != NULL )
      release_graph(&graph_); /*lint !e1551*/
}
#endif



/** frees user data of transformed problem (called when the transformed problem is freed) */
static
SCIP_DECL_PROBDELTRANS(probdeltransTsp)
{
   int i;
   GRAPH* graph;

   graph = (*probdata)->graph;

   for( i = 0; i < graph->nedges; i++ )
   {
      SCIP_CALL( SCIPreleaseVar(scip, &graph->edges[i].var) );
   }
   release_graph(&graph);

   /* free probdata */
   SCIPfreeMemory(scip, probdata);

   return SCIP_OKAY;
}

/**@} */

/**@name Interface methods
 *
 * @{
 */
/** sets up the problem data */
SCIP_RETCODE SCIPprobdataCreate(
   SCIP*                 scip,               /**< SCIP data structure */
   GRAPH*                graph               /**< the graph */
   )
{
   SCIP_PROBDATA* probdata;

   assert(scip != NULL);
   assert(graph != NULL);

   SCIP_CALL( SCIPallocMemory(scip, &probdata) );

   probdata->graph = graph;
   capture_graph(graph);

   /* set user problem data */
   SCIP_CALL( SCIPsetProbData(scip, probdata) );

   /* create a problem in SCIP and add non-NULL callbacks via setter functions */
   SCIP_CALL( SCIPsetProbDelorig(scip, probdelorigTsp) );
   SCIP_CALL( SCIPsetProbTrans(scip, probtransTsp) );
   SCIP_CALL( SCIPsetProbDeltrans(scip, probdeltransTsp) );
   SCIP_CALL( SCIPsetProbCopy(scip, probcopyTsp) );

   return SCIP_OKAY;
}

/** returns the graph */
GRAPH* SCIPprobdataGetGraph(
   SCIP_PROBDATA*        probdata            /**< problem data */
   )
{
   assert(probdata != NULL);

   return probdata->graph;
}

/**@} */
