/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program and library             */
/*         SCIP --- Solving Constraint Integer Programs                      */
/*                                                                           */
/*    Copyright (C) 2002-2015 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SCIP is distributed under the terms of the ZIB Academic License.         */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with SCIP; see the file COPYING. If not email to scip@zib.de.      */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   reader_tsp.c
 * @brief  TSP file reader
 * @author Felipe Serrano
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include <assert.h>
#include <string.h>
#include <ctype.h>

#include "scip/cons_linear.h"
#include "reader_tsp.h"
#include "probdata_tsp.h"
#include "cons_nosubtour.h"
#include "GomoryHuTree.h"


#define READER_NAME             "tspreader"
#define READER_DESC             "tsp file reader"
#define READER_EXTENSION        "tsp"

#define NINT(x) (floor(x+0.5))

/*
 * Local methods
 */

static
SCIP_RETCODE getNodesFromFile(
   SCIP_FILE*         file,               /**< file containing the data to extract */
   double*            x_coords,           /**< double array to be filled with the x-coordinates of the nodes */
   double*            y_coords,           /**< same for y-coordinates */
   GRAPH*             graph               /**< the graph which is to be generated by the nodes */
   )
{
   GRAPHNODE* node = &(graph->nodes[0]);
   char buffer[SCIP_MAXSTRLEN];
   int i;
   int nread;
   int nodenumber;

   /* extract every node out of the filestream */
   i = 0;

   while ( i < graph->nnodes && !SCIPfeof(file) )
   {
      if( SCIPfgets(buffer, (int)sizeof(buffer), file) == NULL )
      {
         printf("invalid node line read\n");
         return SCIP_READERROR;
      }

      /* parse node line */
      nread = sscanf(buffer, "%d %lf %lf\n", &nodenumber, &x_coords[i], &y_coords[i]);
      if( nread < 2 )
      {
         printf("invalid node line in file: %s\n", buffer);
         return SCIP_READERROR;
      }

      /* assign everything  */
      node->id = i;
      if( nodenumber-1 != i)
      {
         printf("warning: nodenumber %d does not match its index in node list %d.\
               Node will get number %d when naming variables and constraints!", nodenumber, i+1, i+1);
      }
      node->x = x_coords[i];
      node->y = y_coords[i];
      node->first_edge = NULL; 
      node++; 
      i++;
   }

   assert( i == graph->nnodes );

   return SCIP_OKAY;
}

/** adds a variable to both halfedges and captures it for usage in the graph */
static
SCIP_RETCODE addVarToEdges(
   SCIP*                 scip,               /**< SCIP data structure */
   GRAPHEDGE*            edge,               /**< an edge of the graph */
   SCIP_VAR*             var                 /**< variable corresponding to that edge */
   )
{
   assert(scip != NULL);
   assert(edge != NULL);
   assert(var != NULL);

   /* add variable to forward edge and capture it for usage in graph */
   edge->var = var;
   SCIP_CALL( SCIPcaptureVar(scip, edge->var) );

   /* two parallel halfedges have the same variable,
    * add variable to backward edge and capture it for usage in graph */
   edge->back->var = edge->var;
   SCIP_CALL( SCIPcaptureVar(scip, edge->back->var) );

   return SCIP_OKAY;
}

/** method asserting that the file had the correct format and everything was set correctly */
static
SCIP_Bool checkValid(
   GRAPH*             graph,              /**< the constructed graph, schould not be NULL */ 
   char*              name,               /**< the name of the file */
   char*              type,               /**< the type of the problem, should be "TSP" */
   char*              edgeweighttype,     /**< type of the edgeweights, should be "EUC_2D", "MAX_2D", "MAN_2D", 
                                           *   "ATT", or "GEO" */
   int                nnodes              /**< dimension of the problem, should at least be one */
   )
{   
   /* if something seems to be strange, it will be reported, that file was not valid */
   if( nnodes < 1 )
   {
      printf("parse error in file %s, dimension should be greater than 0\n", name);
      return FALSE;
   }
   if( strcmp(type, "TSP") != 0 )
   {  
      printf("parse error in file %s, type should be TSP\n", name);
      return FALSE;
   }
   if ( !( strcmp(edgeweighttype ,"EUC_2D") == 0 || strcmp(edgeweighttype, "MAX_2D") == 0 
            || strcmp(edgeweighttype, "MAN_2D") == 0 || strcmp(edgeweighttype, "GEO") == 0 
            || strcmp(edgeweighttype, "ATT") == 0) )
   {
      printf("parse error in file %s, unknown weight type should be EUC_2d, MAX_2D, MAN_2D, ATT or GEO\n", name);
      return FALSE;
   }
   if( graph == NULL)
   {
      printf("parse error in file %s, uninitialized graph, most likely NODE_COORD_SECTION is missing\n", name);
      return FALSE;
   }
   return TRUE;
}

/*
 * Callback methods of reader
 */



/** problem reading method of reader */
static
SCIP_DECL_READERREAD(readerReadTsp)
{  /*lint --e{715}*/
   SCIP_RETCODE retcode;
   SCIP_CONS* cons;

   GRAPH* graph = NULL;
   GRAPHNODE* node;
   GRAPHNODE* nodestart;             /* the two incident nodes of an edge */
   GRAPHNODE* nodeend;
   GRAPHEDGE* edgeforw;              /* two converse halfedges */
   GRAPHEDGE* edgebackw;
   GRAPHEDGE* edge;

   double*  x_coords;
   double*  y_coords;

#ifdef SCIP_DEBUG
   double** weights = NULL;
#endif
   
   double x;                         /* concrete coordinates */
   double y;

   int nnodes = 0;
   int nedges = 0;
   int i;
   int j;

   SCIP_Bool round_lengths;

   SCIP_FILE* file;
   char token[SCIP_MAXSTRLEN];
   char name[SCIP_MAXSTRLEN];
   char type[SCIP_MAXSTRLEN];
   char edgeweighttype[SCIP_MAXSTRLEN];
   char varname[SCIP_MAXSTRLEN];
   char* aux;

   x_coords = NULL;
   y_coords = NULL;

   sprintf(name, "++ uninitialized ++");
   sprintf(type, "TSP");
   sprintf(edgeweighttype, "EUC_2D");

   retcode = SCIP_OKAY;
   *result = SCIP_DIDNOTRUN;

   SCIP_CALL( SCIPgetBoolParam(scip, "reading/tspreader/round_lengths", &round_lengths) );

   /* open file */
   file = SCIPfopen(filename, "r");
   if( file == NULL )
   {
      SCIPerrorMessage("cannot open file <%s> for reading\n", filename);
      SCIPprintSysError(filename);
      return SCIP_NOFILE;
   }

   /* read file */
   if( SCIPfgets(token, (int)sizeof(token), file) == NULL )
   {
      SCIPwarningMessage(scip, "could't read line of file\n");
      return SCIP_READERROR;
   }

   aux = NULL;
   while( !SCIPfeof(file) )
   {
      if( strncmp(token, "NAME", 4) == 0 )
      {
         aux = strchr(token, ':');

         if( aux == NULL )
         {
            SCIPwarningMessage(scip, "did not find problem name\n");
            return SCIP_READERROR;
         }

         /* skip ':' */
         ++aux;

         /* remove white space (tabs, ' ') in front of the name */
         while( isspace((unsigned char)* aux) )
            ++aux;

         /* remove return at end */
         aux[strcspn(aux, "\r\n")] = 0;

         sprintf(name, "%s", aux);
      }
      else if( strncmp(token, "TYPE", 4) == 0 )
      {
         aux = strchr(token, ':');

         if( aux == NULL )
         {
            SCIPwarningMessage(scip, "did not find problem type\n");
            return SCIP_READERROR;
         }

         /* skip ':' */
         ++aux;

         /* remove white space (tabs, ' ') in front of the name */
         while( isspace((unsigned char)* aux) )
            ++aux;

         /* remove return at end */
         aux[strcspn(aux, "\r\n")] = 0;

         sprintf(type, "%s", aux);
      }
      else if( strncmp(token, "DIMENSION", 9) == 0 )
      {
         int nread;

         aux = strchr(token, ':');

         if( aux == NULL )
         {
            SCIPwarningMessage(scip, "did not find problem dimension\n");
            return SCIP_READERROR;
         }

         /* skip ':' */
         ++aux;

         /* remove white space (tabs, ' ') in front of the dimension */
         while( isspace((unsigned char)* aux) )
            ++aux;

         /* remove return at end */
         aux[strcspn(aux, "\r\n")] = 0;

         nread = sscanf(aux, "%d\n", &nnodes);
         if( nread == 0 )
         {
            SCIPwarningMessage(scip, "invalid input for DIMENSION\n");
            return SCIP_READERROR;
         }

         nedges = nnodes * (nnodes-1);
      }
      else if( strncmp(token, "EDGE_WEIGHT_TYPE", 16) == 0 )
      {
         aux = strchr(token, ':');

         if( aux == NULL )
         {
            SCIPwarningMessage(scip, "did not find problem edgeweighttype\n");
            return SCIP_READERROR;
         }

         /* skip ':' */
         ++aux;

         /* remove white space (tabs, ' ') in front of the name */
         while( isspace((unsigned char)* aux) )
            ++aux;

         /* remove return at end */
         aux[strcspn(aux, "\r\n")] = 0;

         sprintf(edgeweighttype, "%s", aux);
      }
      else if( strncmp(token, "NODE_COORD_SECTION", 18) == 0 || strncmp(token, "DISPLAY_DATA_SECTION", 20) == 0 )
      {
         /* there should be some nodes to construct a graph */
         if( nnodes < 1 )
         {
            SCIPwarningMessage(scip, "Not enough nodes: graph is too small. most likely an issue in NODE_COORD_SECTION\n");
            retcode = SCIP_READERROR;
            break;
         }
         /* the graph is created and filled with nodes  */
         else if( create_graph(nnodes, nedges, &graph) )
         {
            SCIP_CALL( SCIPallocBufferArray(scip, &x_coords, nnodes) );
            SCIP_CALL( SCIPallocBufferArray(scip, &y_coords, nnodes) );

            SCIP_CALL( getNodesFromFile(file, x_coords, y_coords, graph) );
         }
         else
         {
            retcode = SCIP_NOMEMORY;
            break;
         }
      }  
      else if( strcmp(token, "\n") == 0 || strncmp(token, "COMMENT", 7) == 0 || strncmp(token, "DISPLAY_DATA_TYPE", 17) == 0 )
         ;
      else if( strncmp(token, "EOF", 3) == 0 )
         break;
      else
      {
         SCIPwarningMessage(scip, "Parse error in file. Unknown keyword: %s\n", token);
         return SCIP_READERROR;
      }
      if( SCIPfgets(token, (int)sizeof(token), file) == NULL )
      {
         SCIPwarningMessage(scip, "Unable to read line of file\n");
         return SCIP_READERROR;
      }
   }/* finished parsing the input */
   
   /* check whether the input data was valid */
   if( !checkValid(graph, name, type, edgeweighttype, nnodes) )
   {
      SCIPwarningMessage(scip, "Invalid data\n");
      retcode = SCIP_READERROR;
   }

   assert(graph != NULL);

   if( retcode == SCIP_OKAY )
   {
      edgeforw = &(graph->edges[0]); /*lint !e613*/
      edgebackw= &(graph->edges[nedges/2]); /*lint !e613*/

#ifdef SCIP_DEBUG
      SCIP_CALL( SCIPallocBufferArray(scip, &weights, nnnodes) );
      for( i = 0; i < nnodes; ++i )
      {
         SCIP_CALL( SCIPallocBufferArray(scip, &(weights[i]), nnnodes) );
      }
#endif

      /* construct all edges in a complete digraph */
      for( i = 0; i < nnodes; i++ )
      {
         nodestart = &graph->nodes[i]; /*lint !e613*/
         for( j = i+1; j < nnodes; j++ )
         {
            nodeend = &graph->nodes[j]; /*lint !e613*/

            /* construct two 'parallel' halfedges */
            edgeforw->adjac = nodeend;
            edgebackw->adjac = nodestart;
            edgeforw->back = edgebackw;
            edgebackw->back = edgeforw;

            /* calculate the Euclidean / Manhattan / Maximum distance of the two nodes */
            x = x_coords[(*nodestart).id] -  x_coords[(*nodeend).id]; /*lint !e613*/
            y = y_coords[(*nodestart).id] -  y_coords[(*nodeend).id]; /*lint !e613*/
            if( strcmp(edgeweighttype, "EUC_2D") == 0 )
               edgeforw->length = sqrt( x*x + y*y );
            else if( strcmp(edgeweighttype, "MAX_2D") == 0 )
               edgeforw->length = MAX( ABS(x), ABS(y) );
            else if( strcmp(edgeweighttype, "MAN_2D") == 0 )
               edgeforw->length = ABS(x) + ABS(y);
            else if( strcmp(edgeweighttype, "ATT") == 0 )
               edgeforw->length = ceil( sqrt( (x*x+y*y)/10.0 ) ); 
            else if( strcmp(edgeweighttype, "GEO") == 0 )
            {
               const double pi =  3.141592653589793;
               double rads[4];
               double coords[4];
               double degs[4];
               double mins[4];
               double euler[3];
               int k;

               coords[0] = x_coords[(*nodestart).id]; /*lint !e613*/
               coords[1] = y_coords[(*nodestart).id]; /*lint !e613*/
               coords[2] = x_coords[(*nodeend).id]; /*lint !e613*/
               coords[3] = y_coords[(*nodeend).id]; /*lint !e613*/

               for( k = 0; k < 4; k++ )
               {
                  degs[k] = coords[k] >= 0 ? floor(coords[k]) : ceil(coords[k]);
                  mins[k] = coords[k] - degs[k];
                  rads[k] = pi*(degs[k]+5.0*mins[k]/3.0)/180.0;
               }
               
               euler[0] = cos(rads[1]-rads[3]);
               euler[1] = cos(rads[0]-rads[2]);
               euler[2] = cos(rads[0]+rads[2]);
               edgeforw->length = floor(6378.388 * acos(0.5*((1.0+euler[0])*euler[1]-(1.0-euler[0])*euler[2]))+1.0);
            }          
            
            /* in TSP community, it is common practice to round lengths to next integer */
            if( round_lengths )
               edgeforw->length = NINT(edgeforw->length);
            
            edgebackw->length = edgeforw->length;
#ifdef SCIP_DEBUG
            weights[i][j] = edgeforw->length;
            weights[j][i] = edgebackw->length;
#endif
   
            /* insert one of the halfedges into the edge list of the node */
            if (nodestart->first_edge == NULL)
            {
               nodestart->first_edge = edgeforw;
               nodestart->first_edge->next = NULL;
            }
            else
            {
               edgeforw->next = nodestart->first_edge;
               nodestart->first_edge = edgeforw;
            }
                   
            /* dito */
            if (nodeend->first_edge == NULL)
            {
               nodeend->first_edge = edgebackw;
               nodeend->first_edge->next = NULL;
            }
            else
            {
               edgebackw->next = nodeend->first_edge;
               nodeend->first_edge = edgebackw;
            }           
                     
            edgeforw++;
            edgebackw++;            
         }
      }
   }

   SCIPfreeBufferArray(scip, &y_coords);
   SCIPfreeBufferArray(scip, &x_coords);
      
   if( retcode != SCIP_OKAY )
   {
#ifdef SCIP_DEBUG
      if( weights != NULL )
      {
         for( i = 0; i < nnodes; i++ )
         {    
            SCIPfreeBufferArray(scip, &weights[i]);
         }
         SCIPfreeBufferArray(scip, &weights);
      }
#endif
      return retcode;
   }

#ifdef SCIP_DEBUG
   printf("Matrix:\n");
   for( i = 0; i < nnodes; i++ )
   {    
      for( j = 0; j < nnodes; j++ )
         printf(" %4.0f ",weights[i][j]);
      printf("\n");
      SCIPfreeBufferArray(scip, &weights[i]);
   }
   SCIPfreeBufferArray(scip, &weights);
#endif

   /* create the problem's data structure */
   /*SCIP_CALL( SCIPcreateObjProb(scip, name.c_str(), new ProbDataTSP(graph), TRUE) ); */
   SCIP_CALL( SCIPcreateProbBasic(scip, name) );
   SCIP_CALL( SCIPprobdataCreate(scip, graph) );

   /* add variables to problem and link them for parallel halfedges */
   for( i = 0; i < nedges/2; i++ )
   {
      SCIP_VAR* var;

      edge = &graph->edges[i]; /*lint !e613*/

      /* the variable is named after the two nodes connected by the edge it represents */
      (void) SCIPsnprintf(varname, SCIP_MAXSTRLEN, "x_e_%d-%d", edge->back->adjac->id + 1, edge->adjac->id + 1);

      SCIP_CALL( SCIPcreateVar(scip, &var, varname, 0.0, 1.0, edge->length,
            SCIP_VARTYPE_BINARY, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL) );

      /* add variable to SCIP and to the graph */
      SCIP_CALL( SCIPaddVar(scip, var) );
      SCIP_CALL( addVarToEdges(scip, edge, var) );

      /* release variable for the reader. */
      SCIP_CALL( SCIPreleaseVar(scip, &var) );
   }

   /* add all n node degree constraints */
   if( nnodes >= 2 )
   {
      char consname[SCIP_MAXSTRLEN];

      for( i = 0, node = &(graph->nodes[0]); i < nnodes; i++, node++ ) /*lint !e613*/
      {
         /* a new degree constraint is created, named after a node */
         (void) SCIPsnprintf(consname, SCIP_MAXSTRLEN, "deg_cons_v%d", node->id + 1);
         SCIP_CALL( SCIPcreateConsLinear(scip, &cons, consname, 0, NULL, NULL, 2.0, 2.0, 
               TRUE, FALSE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE) );  

         edge = node->first_edge;
         /* sum up the values of all adjacent edges  */
         while( edge != NULL )
         {
            SCIP_CALL( SCIPaddCoefLinear(scip, cons, edge->var, 1.0) );
            edge = edge->next;
         }
             
         /* add the constraint to SCIP */
         SCIP_CALL( SCIPaddCons(scip, cons) );
         SCIP_CALL( SCIPreleaseCons(scip, &cons) );
      }
   }

   /* finally, we need a constraint forbidding subtours */
   SCIP_CALL( SCIPcreateConsNosubtour(scip, &cons, "nosubtour", graph, 
            FALSE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE) );
   SCIP_CALL( SCIPaddCons(scip, cons) );
   SCIP_CALL( SCIPreleaseCons(scip, &cons) );
 
   release_graph(&graph);
   *result = SCIP_SUCCESS;

   return SCIP_OKAY;
}


/** problem writing method of reader */
static
SCIP_DECL_READERWRITE(readerWriteTsp)
{  /*lint --e{715}*/
   *result = SCIP_DIDNOTRUN;

   return SCIP_OKAY;
}


/*
 * reader specific interface methods
 */

/** includes the tsp file reader in SCIP */
SCIP_RETCODE SCIPincludeReaderTsp(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_READERDATA* readerdata;
   SCIP_READER* reader;

   /* create reader data */
   readerdata = NULL;

   /* include reader */
   SCIP_CALL( SCIPincludeReaderBasic(scip, &reader, READER_NAME, READER_DESC, READER_EXTENSION, readerdata) );

   assert(reader != NULL);

   /* set non fundamental callbacks via setter functions */
   SCIP_CALL( SCIPsetReaderRead(scip, reader, readerReadTsp) );
   SCIP_CALL( SCIPsetReaderWrite(scip, reader, readerWriteTsp) );

   /* add tsp reader parameters */
   SCIP_CALL( SCIPaddBoolParam(scip,
         "reading/tspreader/round_lengths", "should lenghts of edges be rounded to nearest integer?",
            NULL, FALSE, TRUE, NULL, NULL) );

   return SCIP_OKAY;
}
